(*
 * LANGUAGE	: ANS Forth
 * PROJECT	: Forth Environments
 * DESCRIPTION	: Dynamic linking to OS libraries
 * CATEGORY	: Utilities
 *		: 2009-08-10 23:33:28, fixed bug when importing a library function again
		: 2009-08-12 05:03:08, forgetting corrected
 * LAST CHANGE 	: Sunday, August 09, 2009, 17:25, Marcel Hendrix, review and cosmetics
		: 2009-08-23 09:23:58;  forgetting and AtExit both correct
		: 2009-08-25 00:16:58;  added support for )C-long functions
		: 2009-08-25 19:44:08;  redesign allowing 28 mixed-type parameters; see GET-ONE-PARAMETER
		: 2009-08-25 19:44:58,  added debugging support
		: 2009-09-03 16:24:15;  toplib info removed exept for debugging as in discussion
		: 2009-09-03 21:18:21;  added support for LOCATE
		: 2009-09-04 23:32:56;  redefinitions warnings; added AliasedLibrary:
		: 2009-09-20 01:02:07;  added Extern: extern function name ; no library call but only addr
		: 2009-09-22 23:28:09;  fixed evaluation bug; .externs better information
		: 2009-09-25 19:48:50;  changed the linked list for functions to test for the newest correct mode,
					the @LATEST field is corrected to be either the name field in libraries
					as they can be forgotten without problems, or the body for functions to
					test the address in dictionary space
		: 2009-10-14 14:12:01	User Reference Manual, added library-removing flag, rewritten test code
 * AUTHOR      : Hanno Schwalm
 *)


	NEEDS -miscutil

	REVISION -dynlibs " OS Library linking  Version 1.22 "

(*	***** The User Reference Manual *****

	*** What is dynlibs used for?
	In many situations it is conveniant to use libraries developed by other persons and
	written in other languages in iForth applications - noone like reinventing wheels.
	In this scenario you have decided to use a library, in this example we will use the
	freeglut library. This is all tested in deep and it's the basis of already existing iForth
	extensions like fjack and the new opengl system.
	BUT - there might still be bugs
	    - the extern: systax has safety restrictions.
	If you run in trouble don't hesitate to contact me or marcel.

	*** Module options
	To give you more ways to control what dynlib does you might set one of the debugging flags
	to TRUE - you'll find them right after this manual.
	There are two more options,
	SEARCH-ALL-LIBRARIES?	Can be set to FALSE when you only want top search in the topLib
				if TRUE all open libraries are searched
	PROHIBIT-CLOSE-LIB?	Is normally TRUE, this means that libraries are closed when the
				application is left or by FORGET or MARKER execution
				Some libraries might have problems with closing, in this scenario
				you can set the flag to FALSE.


	Also remember what the linker of dynamic libraries does; it may well load more libraries -
	those the loaded library depend on.

	Each library has a lot of built-in functions, these functions use parameters in a different
	way forth uses them. For each of the functions who will have to tell the parameters in a
	C-like way. To make life easy you can copy&paste the function definitions in the .h file of
	the library and use a Forth word Extern: to anylyse this description. Extern: compiles a standard
	colon definition which can be used as all other definitions.
	Some other important points; loading the libraries and finding the functions address in the
	library is done
		a) at compile time so you can use the word atonce as all other colon definitions
		b) The libraries defined in the application are loaded also when starting your
		   turnkey application befor any application code is executed, this is done in the cold-chain
		c) Also the funtions are made valid in turnkey applications by searching for their
		   address in the cold chain.
		d) When a function is not found - the library might simply no installed on your system or
		   the installed version of the library does not have this function - your code will not
		   crash but an ABORT" message is displayed.
	You may inspect the status of the defined libraries and functions at all times.
	dynlibs can be used with every recent iForth version 4.0 or later on all supported platforms,
	it also supports the LOCATE mechanism. Of course FORGET and MARKER are also recognised.


	*** Defining the Libraries
		There are two word to define a library
	Library: name
		This expects the real name of the libary as the system names it, the freeglut library would be
		Library: libglut.so
	LIBRARY is an alias for Library:

		You can also define an alias name for the library for better readability or when you write
		applications for different targets
	AliasedLibrary: name real-library-name
		AliasedLibrary: glut-library libglut.so
		In Forth the name now would be glut-library instead of libglut.so. In both cases
		the defined library-word is a normal forth word that can be executed.
		It returns the handle of the library or 0 if not found.

		Sometimes a function name might be found in several libraries. Normally all libraries are
		searched for a function, you may specify the first search by
	topLib ( library-handle -- )

		To see the library status use
	.LIBRARIES ( -- )


	*** Defining the functions
		There are basically two word to define a function in a library; please remember defining
		a function means compiling a colon definition - nothing else from your point. Internally
		registers are saved and later restored, parameters are transferred from the Forth style
		- integer parameters from the normal stack and floats of course from the floating stack -
		to the C-style parameters. This uses the really complicated iForth internals wisdom about
		how these parameters are presented to the C-function depending on the underlying operating
		system.
	.EXTERNS ( -- ) Will show you all defined Extern: definitions.

		A first example to define a function looks like
	Extern: void glMaterialfv ( GLenum face, GLenum pname, const GLfloat * params );
		^^^^ defines the type of result. Accepted types are
			void	no result
			float
			double
			long	2 cell integer
			int
			char	Same as integer
			extern	Very different - this is not a normal result! It will return the address
				in the library when the function is defined with a & in the library.
				See for the glutfont examples in glutlib and the library header to understand
				what it's about.
			const	and
			*	are both dummys they MUST be followed by one of the above function types.
		When none of these types is given the Extern: compiler assumes the result of the function is
		an integer. Take care this assumption is met, there might be some #defines in the library
		that makes this incorrect! Especially look out for floating results and change accordingly!

		There is an addition here to make this Extern: compatible with the VFX way. There might be
		an additional 'sort-type-instruction' - these are not used in iForth environments. Allowed
		'extras' are "C" PASCAL and "PASCAL".

	Extern: void glMaterialfv ( GLenum face, GLenum pname, const GLfloat * params );
		     ^^^^^^^^^^^^ This is the exact name of the function in the library; please make sure
			there are no typos or case problems. You won't notice this as usual when writing forth
			code as this is also case-sensitive and there is NO way to assume a correct spelling.
			Also the Extern: definition will be compiled without a problem - BUT when you execute
			it it will abort! Why this? As the function with the wrong name is searched and not
			found and this will lead to the ABORT message "function not loaded". You may look
			for this problem by .EXTERN, normally the function addresses are displayed, if you
			see a 0 this means this functions could not be found.

	Extern: void glMaterialfv ( GLenum face, GLenum pname, const GLfloat * params );
				 ^^^ A space seperated '(' follows, now the parameters for the function are
			declared. At the end of the parameter list is a closing ')' followed by ';' as this
			is the C Syntax.
			The Extern: analyser scans the string between the ( ) pair and seperates it to the
			subparts seperated by the ','. Every substring will be one parameter, if none is found
			the parameter for the function will be defaulted to void.
			Each paraneter string is analysed by this rule:
			1) If there is a '*' detected this must be a pointer, for forth it's an integer.
			2) a void is a void :-)
			3) float and double are floating point parameters taken from the floating point
			   stack.
			4) ALL others are must be integers. (When the iForth C( syntax accepts oher types as
			   long-integers this will be added.)
			That's it. The original in gl.h looked like
		WINGDIAPI void APIENTRY glMaterialfv (GLenum face, GLenum pname, const GLfloat *params);
			so not much had to be changed.


		In Extern: definitions the forth word is exactly the same as the library function name.
		You can also use alias definitions like
	AliasedExtern: rotate! void glRotated ( double angle, double x, double y, double z);
		       ^^^^^^^ it's pretty much the same as an Extern: definition execpt the forth name is
		       following right after AliasExtern:, the rest is exactly the same.

		You might want to know where a library function is located or when it could not be found.
	FUNC-LOADED? ( xt -- code-address|0 )
		will do this for you. When giving away applications it's good style to test for a function
		in the library and give a proper error message when the program doesn't work because the
		library is missing or doesn't have the function included because of a version problem.


		*** The old style library interface
		There were library functions available before dynlibs made it into light, the 'old'
		syntax used by some examples are supported via these:
	IMPORT  ( exec: #params "name" -- ) ( comp: #params*{int} -- ior )
	IMPORTF ( exec: #params "name" -- ) ( comp: #params*{int} -- ior )
	IMPORTV ( exec: #params "name" -- ) ( comp: #params*{int} -- )

		*** Practice! ***
		When you want to go ahead with your own application i suggest to look into three 'eamples'
		1) include/glutlib.frt declares libraries and functions; it's the first step on the way
		to the new portable iForth opengl interface
		2) include/fjack/jacklib.frt defines functions used by fjack. Please note that this
		   file is also used by VFX so there are the "C" instructions simply ignored in iForth.
		3) include/fjack/I_osdeps.frt shows the declaration of the libraries for fjack.



*)

PRIVATES


\ ***************** general Options when loading this module
\ If TRUE other libraries are searched AFTER the current with a warning when found somewhere else
TRUE CONSTANT SEARCH-ALL-LIBRARIES? PRIVATE
\ if TRUE libraries are not automatically removed
FALSE CONSTANT PROHIBIT-CLOSE-LIB? PRIVATE



\ This debugging flag can be set top true for tracking the function parser/generator
FALSE  \ DROP TRUE
[IF]	: DEBUG| ; IMMEDIATE PRIVATE
	: | ; IMMEDIATE PRIVATE
[ELSE]	: DEBUG| [CHAR] | WORD DROP ; PRIVATE IMMEDIATE
[THEN]

\ This debugging flag can be set to true when you want debugging output in library open/search
FALSE  \ DROP TRUE
[IF]	: SDEBUG| ; IMMEDIATE PRIVATE
	[UNDEFINED] | [IF]  : | ; IMMEDIATE PRIVATE [THEN]
[ELSE]	: SDEBUG| [CHAR] | WORD DROP ; PRIVATE IMMEDIATE
[THEN]


DOC
(*
	The library struct
	------------------
		0 points to library filename
		1 the library handle
		2 link to the next library
		3 points to forth representation name of library or 0 when not an alias

	The extern struct
	-----------------
		0 pointer to nfa of Forth word; may be alias
		1 holds function address
		2 library handle
		3 function type
		4 number of parameters (parameter type is assumed to be always int)
		5 link to the next extern struct
		6 pointer to the function's name in the library
*)
ENDDOC


CREATE FIRST-EXTERN 0 , 0 , 0 , 0 , 0 , PRIVATE	\ the dummy lowest function node
: TOP-EXTERN	( -- addr ) \ returns the topmost functionnode in dictionary space
	FIRST-EXTERN
	BEGIN	DUP 5 CELLS + @ FIRST-EXTERN HERE WITHIN
		IF	DUP 5 CELLS + @ @ FIRST-EXTERN HERE WITHIN
		ELSE	FALSE
		THEN
	WHILE	5 CELLS + @
	REPEAT ; PRIVATE

0 VALUE LINK-LIBRARY   PRIVATE
0 VALUE CURRENT-LIBRARY PRIVATE

: add-mode-to-pars	( pars mode -- pars )
	OVER #15 AND #13 >	ABORT" too many parameters for library function"
	DUP 0 3 WITHIN 0=	ABORT" incorrect parameter for library function"
	4 LSHIFT SWAP DUP $0F AND 1+ SWAP $FFFFFFF0 AND 2 LSHIFT OR OR ; PRIVATE
: get-nth-mode		( pars n -- )
	OVER $0F AND SWAP - 2* 2+ RSHIFT 3 AND ; PRIVATE
: +-trailing
	BEGIN	2DUP 0> SWAP C@ DUP BL <= SWAP '(' = OR AND
	WHILE	1- SWAP 1+ SWAP
	REPEAT
	BEGIN	2DUP DUP 0> -ROT + 1- C@ DUP BL <= SWAP ')' = OR AND
	WHILE	1-
	REPEAT ; PRIVATE	\ ending characters removed

: (.Extern)	( struct )
	LOCAL struct
	CR SDEBUG| struct 2 CELLS + @ ?DUP IF @ ." (topLib " .$ ." ) " THEN |
	struct 1 CELLS + @ H. ?AT DROP 10 MOD 10 SWAP - 1 MAX SPACES
	struct 3 CELLS + @ ?DUP 0= IF EXIT THEN
	CASE	1 OF	."    int "	ENDOF
		2 OF	."   void "	ENDOF
		3 OF	."  float "	ENDOF
		4 OF	." double "	ENDOF
		5 OF	."   long "	ENDOF
		6 OF    ." extern &"	ENDOF
	ENDCASE
	struct 6 CELLS + .$
	struct 3 CELLS + @ 6 <>
	IF	 ."  ( " struct 4 CELLS + @ DUP $0F AND 0
		?DO	DUP I get-nth-mode
			CASE	0 OF	." _void "	ENDOF
				1 OF	." _int "	ENDOF
				2 OF	." _double "	ENDOF
			ENDCASE
		LOOP	DROP ." );"
	THEN
	struct @ >HEAD ?DUP 0= ?EXIT
	HEAD>NAME COUNT 2DUP struct 6 CELLS + COUNT COMPARE IF ."   --> " TYPE ELSE 2DROP THEN
	  ; PRIVATE
: (.LIBRARY)	( struct -- )
	DUP @ .$ ."  (" DUP 1 CELLS + @ H. ." )"
	3 CELLS + @ ?DUP IF ."   --> " .$ THEN  ; PRIVATE

: LIBRARY-ERROR 0 #163 SYSCALL ;      PRIVATE	( -- c-addr u )
: LIBRARY-OPEN			( c-addr u -- handle ior )
	SDEBUG| ."  open library " |
 	2 #160 SYSCALL
	SDEBUG| DUP IF ."  --- not found " ELSE ." , handle is " OVER H. THEN |
	;      PRIVATE
: LIBRARY-CLOSE
	SDEBUG| CR ." library (handle): " DUP H. |
	PROHIBIT-CLOSE-LIB?
	IF	SDEBUG| ."  still there" | EXIT THEN
	1 #161 SYSCALL
	SDEBUG| ."  closed" DUP IF ." with error" THEN  |
	DROP ; PRIVATE
: LIBRARY-FIND			( c-addr u handle -- addr ior )
	SDEBUG| ( ." handle# " DUP H. ) ."  searching for " >R 2DUP TYPE R> |
 	3 #162 SYSCALL
	SDEBUG| DUP IF ."  not found" ELSE ."  found at " OVER H. THEN | ; PRIVATE

: .Externs ( -- )
	FIRST-EXTERN 5 CELLS + @
	BEGIN	DUP FIRST-EXTERN TOP-EXTERN 1+ WITHIN
	WHILE	DUP (.Extern)
		5 CELLS + @
	REPEAT	DROP ;
: ShowExtern ( -- ) \ name
	BL <WORD> LOCALS| cnt addr |
	FIRST-EXTERN 5 CELLS + @
	BEGIN	DUP FIRST-EXTERN TOP-EXTERN 1+ WITHIN
	WHILE	DUP @ COUNT addr cnt COMPARE 0=
		IF DUP (.Extern) THEN
		5 CELLS + @
	REPEAT	DROP ;
: .FUNCTIONS ( -- )	.Externs ;
: .LIBRARIES ( -- )	LINK-LIBRARY BEGIN DUP WHILE DUP CR (.LIBRARY) 2 CELLS + @ REPEAT DROP ;

: SET-FUNCTION-ADDRESS	( struct -- )	\ assumes the library is already loaded
	LOCAL struct
	struct 2 CELLS + @ 0= ABORT" No current library set"
	struct 1 CELLS + OFF
	struct 2 CELLS + @
	1 CELLS + @ ?DUP
	IF	SDEBUG| CR ." In " struct 2 CELLS + @ @ .$ SPACE |
		struct 6 CELLS + COUNT ROT LIBRARY-FIND
		IF DROP	ELSE struct 1 CELLS + !	THEN
	THEN ; PRIVATE
: SEARCH-SET-FUNCTION-ADDRESS	( struct -- )
	LOCAL struct
	struct SET-FUNCTION-ADDRESS
	SEARCH-ALL-LIBRARIES? 0=  struct 1 CELLS + @ 0<> OR IF EXIT THEN
	SDEBUG| struct 1 CELLS + @ 0= IF CR ." Warning: " struct 6 CELLS + .$ ."  not found in " struct 2 CELLS + @ @ .$ SPACE THEN |
	LINK-LIBRARY
 	BEGIN	DUP
 	WHILE	DUP struct 2 CELLS + ! struct SET-FUNCTION-ADDRESS
		struct 1 CELLS + @ 0<>
		IF	SDEBUG| SPACE struct 2 CELLS + @ @ .$ |
			DROP EXIT
		THEN
		2 CELLS + @
	REPEAT	DROP ; PRIVATE
:NONAME	( -- ) \ LOAD-ALL-LIBRARIES & FUNCTIONS
	LINK-LIBRARY
	BEGIN	DUP
	WHILE	SDEBUG| CR ." Load library: " DUP @ COUNT TYPE |
		DUP @ COUNT LIBRARY-OPEN IF SDEBUG| ."  missing" | DROP 0 THEN	( link handle|0 )
		OVER 1 CELLS + !
		2 CELLS + @
	REPEAT	DROP
	FIRST-EXTERN 5 CELLS + @
	BEGIN	DUP FIRST-EXTERN TOP-EXTERN 1+ WITHIN
	WHILE	DUP SET-FUNCTION-ADDRESS 5 CELLS + @
	REPEAT	DROP ; AtCold

: CLOSE-ALL-LIBRARIES	( -- )	\ CLOSE-ALL-LIBRARIES
	LINK-LIBRARY	( close all libraries in list )
	BEGIN	DUP
	WHILE	DUP 1 CELLS + @ ?DUP IF LIBRARY-CLOSE DROP THEN
		2 CELLS + @
	REPEAT	DROP ; PRIVATE ' CLOSE-ALL-LIBRARIES AtExit


\ Isolate the function type definer
\ Maybe there is a VFX forth dummy declaring the stack conversion mode either "C" "PASCAL" OR PASCAL
\ iForth implements stack ordering in the )C-xxx
\ if not so -> the word was the name and will be read again
: REMOVE-VFX-ORDERS	( -- )
	>IN @ BL <WORD>	 \ pos addr cnt
	2DUP S" PASCAL"   COMPARE 0= IF 2DROP DROP EXIT THEN
	2DUP S~ "PASCAL"~ COMPARE 0= IF 2DROP DROP EXIT THEN
	     S~ "C"~      COMPARE 0=
	IF DROP ELSE >IN ! THEN ; PRIVATE

\ reads the function type from the input stream
: GET-EXTERN-TYPE	( -- type )	\ 1 int; 2 void; 3 float; 4 double;  5 long-int;)
	0 LOCAL functype	\ integer is default but undefined yet for safer testing
	BL <WORD>
	2DUP S" const"	COMPARE 0= IF 2DROP BL <WORD>	THEN
	2DUP S" *"	COMPARE 0= IF 2DROP BL <WORD>	THEN
	2DUP S" void"	COMPARE 0= IF 2 TO functype	THEN
	2DUP S" float"	COMPARE 0= IF 3 TO functype	THEN
	2DUP S" double"	COMPARE 0= IF 4 TO functype	THEN
	2DUP S" long"	COMPARE 0= IF 5 TO functype	THEN
	2DUP S" int"	COMPARE 0= IF 1 TO functype	THEN
	2DUP S" char"	COMPARE 0= IF 1 TO functype	THEN
	2DUP S" extern"	COMPARE 0= IF 6 TO functype	THEN
	DEBUG| functype 0= IF	CR ." Warning: function return value (" 2DUP TYPE ." ) set to int, please make sure" THEN |
	2DROP functype 1 MAX
	REMOVE-VFX-ORDERS ; PRIVATE


\ Getting-one-parameter follow this strategy
\ 1) it splits the first part from the string, the separator is ','
\ 2) when no separator is found the whole string is used
\ 3) now the seperated first part is made as short as possible by removing blanks at start/end
\ 4) a nullstring must be a void
\ 5) a void is a void also
\ 6) if there is a '*' somewhere it must be a pointer, for forth it means an integer
\ 7) float and double are both accepted as a floating parameter;
\ 8) all others are something else so here they are integers

: GET-ONE-PARAMETER	( par addr cnt -- addr' cnt' par )
	OVER 0	( addr' cnt' addr cnt )
	BEGIN	2OVER 0> SWAP C@ ',' <> AND	\ scan for a , character
	WHILE	1+ 2>R 1- SWAP 1+ SWAP 2R>	\ make the first part longer; the second shorted until the , is found
	REPEAT					\ a leading , in the second part is removed at once
	2>R 2DUP 0> SWAP C@ ',' = AND IF 1- SWAP 1+ SWAP THEN 2R>
	+-trailing				\ make the first part as short as possible
						\ a null-string must be a void -> give warning
	DEBUG| DUP 0= IF CR ." Warning: probably missing void in " SOURCEFILENAME TYPE ."  #line: " SOURCELINE# . 1000 MS THEN |

	\ now there are both strings on the stack; on to is the one to be checked now
	( pars' addr' cnt' addr cnt )					DEBUG| CR 2DUP TYPE ." --> " |

	\ test for very short string or void first
	2DUP S" void" COMPARE 0= OVER 2 < OR
	IF	\ when a condition is true as here, the checked string is removed before exit
		\ also the 'mode' is rotated to the mask. The lower 4bits in the mask work as a counter
		2DROP ROT 0 add-mode-to-pars -ROT DEBUG| ." void" | EXIT
	THEN

	\ now check for a pointer
	2DUP BOUNDS
	DO	I C@ [CHAR] * =
		IF	DROP 0						DEBUG| ." pointer is int" |
			\ a pointer won#t be interpreted any longer, so length is set to 0
			2>R ROT 1 add-mode-to-pars -ROT 2R> LEAVE
		THEN
	LOOP
	DUP 0= IF 2DROP EXIT THEN					\ was a pointer so no more type checking

	\ floats and doubles probably both work as double in the C( -- )C sense
	OVER 6 S" float " COMPARE 0= >R OVER 7 S" double " COMPARE 0= R> OR
	IF	2DROP ROT 2 add-mode-to-pars -ROT DEBUG| ." float" | EXIT
	THEN

	\ when no void or float was fount the current parameter MUST be an integer
	DEBUG| ." default is int" |
	2DROP ROT 1 add-mode-to-pars -ROT ; PRIVATE

: GET-PARAMETER-CNT	( -- n )
\ Find the parameter string upto the closing ")", removing meaningless characters
	[CHAR] ; <WORD> +-trailing					DEBUG| CR CR ." parameters: " 2DUP TYPE |
	0 -ROT BEGIN GET-ONE-PARAMETER DUP 0<= UNTIL 2DROP ; PRIVATE

\ Takes parameters from the function description struct and makes a string that
\ can be evaluated to make the new Forth word.
\ The name of the word is ( name len )
0 VALUE ->EXTERN	\ dummy to pass a parameter
: COMPILE-FORTH-EXTERN	( name len struct -- )
	$100 ALLOCATE THROW LOCALS| buff struct |
	S" : " buff PACK DROP  buff PLACE+
	struct 1 CELLS + TO ->EXTERN			\ pointer to addr in library
	struct 3 CELLS + @ 6 =
	IF	S~  [ ->EXTERN ] LITERAL @ ;~ buff PLACE+
		\ S~  [ ->EXTERN ] LITERAL @ DUP 0= ABORT" No extern definition found" ;~ buff PLACE+
	ELSE	S"  [ ->EXTERN ] LITERAL @ DUP IF C( " buff PLACE+
		\ here the number of type of parameters is coded; low 4 bits are count; above are 2bit pairs for each parameter
		struct 4 CELLS + @ DUP $0F AND 0 \ number of parameters
		?DO	DUP I get-nth-mode
			CASE	0 OF	S"  _void " buff PLACE+		ENDOF
				1 OF	S"  _int " buff PLACE+		ENDOF
				2 OF	S"  _double " buff PLACE+	ENDOF
			ENDCASE
		LOOP	DROP
		struct 3 CELLS + @  \ type
		CASE	1 OF	S" )C-int "    buff PLACE+	ENDOF
			2 OF	S" )C-void "   buff PLACE+	ENDOF
			3 OF	S" )C-float "  buff PLACE+	ENDOF
			4 OF	S" )C-double " buff PLACE+	ENDOF
			5 OF	S" )C-long "   buff PLACE+	ENDOF
			TRUE ABORT" COMPILE-FORTH-EXTERN :: unknown/unsupported type identifier"
		ENDCASE
		S~  ELSE TRUE ABORT" library function not available" THEN ;~ buff PLACE+
	THEN
	buff COUNT DEBUG| CR 2DUP TYPE | EVALUATE
\ Patch the name of the function into the struct
	@LATEST HEAD> @ struct !	\ not the dea but the execution token
	SET-LATEST-LOCATION
	buff FREE DROP ; PRIVATE

: MAKE-EXTERN-STRUCT	( -- struct )
	ALIGN HERE >R 6 CELLS ALLOT R@ 6 CELLS ERASE
	\ Keep the external function's name available after the struct
	BL <WORD> HERE OVER 2 + ALLOT PACK DROP	R> ; PRIVATE

: .REDEFINITION	( -- )	BL <WORD> TYPE ." --> WARNING is not compiled beacause it already exists" ; PRIVATE

( exec: #params float? "name" -- ) ( comp: #params*{int} float? -- ior )
: (IMPORT)
	0 LOCALS| struct func-type parameters |
	>IN @ ?DEF SWAP >IN !					\ forget it when already defined
	IF	CR ." (IMPORT) " .REDEFINITION EXIT
	THEN
	MAKE-EXTERN-STRUCT TO struct
	CURRENT-LIBRARY		struct 2 CELLS + !
	struct SEARCH-SET-FUNCTION-ADDRESS
	func-type      		struct 3 CELLS + !		\ function type
	0 parameters 0 DO 1 add-mode-to-pars LOOP
				struct 4 CELLS + !		\ parameter matrix in one cell
	struct TOP-EXTERN	5 CELLS + !
	struct 6 CELLS + COUNT struct COMPILE-FORTH-EXTERN ;	\ compile the word
	PRIVATE

\ ******************************* PUBLIC words *****************

\ Extern: jack_port_t "C" jack_port_register ( jack_client_t * client, const char * port_name, const char * port_type, int flags, int buffer_size);
\ Builds an extern struct and the corresponding Forth word. It scans the input line so parts must be separated by whitespace.
\ When the function has already been declared it's not done again.
\ a) for AliasExterns the Forth word name follows
\ b) the function type, if the type is not "float" or "void" it's assumed int
\ c) to be compatible with the VFX Forth, the dummies "C", "PASCAL", and PASCAL are ignored
\ d) the name of the library function
\ e) a "(" to start the parameters
\ f) parameters separated by ","
\ g) the ending ")" and a following ";"

\ Externs will have zero as address when not available in library; this can be tested by FUNC-LOADED?
: Extern:
	0 0 LOCALS| struct func-type |
	GET-EXTERN-TYPE TO func-type
	>IN @ ?DEF SWAP >IN !					\ forget it when already defined
	IF	CR ." Extern: " .REDEFINITION [CHAR] ; <WORD> 2DROP EXIT
	THEN
	MAKE-EXTERN-STRUCT TO struct
	SDEBUG| CR ." Make Extern: " struct 6 CELLS + COUNT TYPE |
	CURRENT-LIBRARY		struct 2 CELLS + !
	struct SEARCH-SET-FUNCTION-ADDRESS
	func-type 	 	struct 3 CELLS + !		\ function type
	GET-PARAMETER-CNT 	struct 4 CELLS + !		\ parameter matrix in one cell
	struct TOP-EXTERN	5 CELLS + !
	struct 6 CELLS + COUNT  struct COMPILE-FORTH-EXTERN ;	\ compile the word

: AliasedExtern:
	0 0 LOCALS| struct func-type |
	>IN @ ?DEF SWAP >IN !		\ forget it when already defined
	IF	CR ." AliasedExtern: " .REDEFINITION [CHAR] ; <WORD> 2DROP EXIT
	THEN
	BL <WORD> 2>R	\ The alias name has to be presented to the evaluate string, so get it now
	GET-EXTERN-TYPE TO func-type
	MAKE-EXTERN-STRUCT TO struct
	CURRENT-LIBRARY		struct 2 CELLS + !
	struct SEARCH-SET-FUNCTION-ADDRESS
	func-type 		struct 3 CELLS + !		\ function type
	GET-PARAMETER-CNT 	struct 4 CELLS + !		\ parameter matrix in one cell
	struct TOP-EXTERN	5 CELLS + !
	2R> struct COMPILE-FORTH-EXTERN ;			\ compile the word

: IMPORT  ( exec: #params "name" -- ) ( comp: #params*{int} -- ior ) 1 (IMPORT) ;
: IMPORTF ( exec: #params "name" -- ) ( comp: #params*{int} -- ior ) 3 (IMPORT) ;
: IMPORTV ( exec: #params "name" -- ) ( comp: #params*{int} -- )     2 (IMPORT) ;

\ Define a library with name, it is available at once and turnkey applications can use it.
\ The runtime code of the defined word returns the handle if available or 0 when not available.
: Library:  \ ( -- ) name
	>IN @ ?DEF SWAP >IN !
	IF	CR ." Library: " .REDEFINITION EXIT
	ENDIF
	CREATE	@LATEST HEAD>NAME DUP , COUNT SDEBUG| CR 2DUP TYPE |
		LIBRARY-OPEN IF DROP 0 THEN	,
		LINK-LIBRARY HERE 2 CELLS - DUP TO CURRENT-LIBRARY TO LINK-LIBRARY , 0 ,
 	FORGET>	LINK-LIBRARY	( addr link )
 		BEGIN	2DUP <=
 		WHILE	DUP 1 CELLS + @ LIBRARY-CLOSE DROP
 			2 CELLS + @
 		REPEAT	NIP TO LINK-LIBRARY
		0 TO CURRENT-LIBRARY
	DOES>	1 CELLS + @ ;
: AliasedLibrary:  \ ( -- ) name real-library-name
	>IN @ ?DEF SWAP >IN !
	IF	CR ." AliasedLibrary: " .REDEFINITION BL <WORD> EXIT
	ENDIF
	CREATE	SDEBUG| CR @LATEST HEAD>NAME COUNT TYPE ." -->" |
		HERE 4 CELLS + ,		\ the real filename is written later, here the pointer is set
		BL <WORD> SDEBUG| 2DUP TYPE |
		2DUP LIBRARY-OPEN IF DROP 0 THEN ,
		LINK-LIBRARY HERE 2 CELLS - DUP TO CURRENT-LIBRARY TO LINK-LIBRARY ,
		@LATEST HEAD>NAME ,
		HERE OVER 2+ ALLOT PACK DROP
	FORGET>	LINK-LIBRARY	( addr link )
 		BEGIN	2DUP <=
 		WHILE	DUP 1 CELLS + @ LIBRARY-CLOSE DROP
 			2 CELLS + @
 		REPEAT	NIP TO LINK-LIBRARY
		0 TO CURRENT-LIBRARY
	DOES>	1 CELLS + @ ;

: LIBRARY	Library: ; ( "name" -- ) \ e.g. winconst.frt needs this.
: topLib	( handle -- )
	LOCAL handle
	LINK-LIBRARY
	BEGIN	DUP
	WHILE	DUP 1 CELLS + @  handle =
		IF	DUP TO CURRENT-LIBRARY
			SDEBUG| CR ." Current Library set to " CURRENT-LIBRARY (.LIBRARY)
			DROP EXIT
		THEN
		2 CELLS + @
	REPEAT	DROP
	TRUE ABORT" Couldn's set the current library" ;

\ This tool allows to find out at runtime whether a function is available or not.
\ Helps to write code that tests for library "versions".
:NONAME	DROP CLOSE-ALL-LIBRARIES ;
: FUNC-LOADED?	( xt -- code-address|0 )
	FIRST-EXTERN 5 CELLS + @
	BEGIN	DUP FIRST-EXTERN TOP-EXTERN 1+ WITHIN
	WHILE	2DUP @ = IF NIP 1 CELLS + @ EXIT THEN
		5 CELLS + @
	REPEAT	2DROP 0 ;
	IS-FORGET FUNC-LOADED?

:ABOUT	CR ." ** Link dynamically to C functions **" ;


NESTING @ 1 =
  [IF]	MARKER test	\ Some ugly test code will be forgotten right now

	LINUX? [IF]
		Library: libc.so.6
		AliasedExtern: labs int labs ( int n );
		: TEST-LIB
			CR ." Executing abs(-2) = " -2 labs .
			CR ." Executing abs( 2) = "  2 labs . ;
	[THEN]

	DARWIN? [IF]
		Library: libc.dylib
		AliasedExtern: labs int abs ( int n );
		: TEST-LIB
			CR ." Executing abs(-2) = " -2 labs .
			CR ." Executing abs( 2) = "  2 labs . ;
	[THEN]

	WINDOWS? [IF]
		(*
		  The GetNumberOfConsoleMouseButtons function retrieves the number of buttons
		  on the mouse used by the current console.

		  BOOL GetNumberOfConsoleMouseButtons(
		      LPDWORD  lpNumberOfMouseButtons 	// address of number of buttons
		   );

		  Parameters
		   lpNumberOfMouseButtons
		   Points to a 32-bit variable that receives the number of mouse buttons.
		  Return Value
		   If the function succeeds, the return value is TRUE.
		   If the function fails, the return value is FALSE. To get extended error
		   information, call GetLastError.
		*)
		Library: kernel32.dll
		Extern: int GetNumberOfConsoleMouseButtons ( * buffer );
		VARIABLE buff
		: TEST-LIB ( -- )
			CR ." reading number of mouse buttons"
			buff GetNumberOfConsoleMouseButtons IF ."  problem" THEN
			." , the number is: " buff 32B@ H. ;

	[THEN]
	CR CR .( known libraries) .LIBRARIES
	CR CR .( functions included) .EXTERNS
	CR CR .( execute some test code) TEST-LIB
	test
[THEN]



DOC
-- Let's test the overhead.
0   VALUE dll		PRIVATE
0   VALUE '#buttons?	PRIVATE
0   VALUE 'GTC		PRIVATE
0. DVALUE 1e6calls	PRIVATE

64BIT?
  [IF]	CODE dummy ( addr -- u )
	     $48 C, $C7 C, $C0 C,  1 C, 0 C, 0 C, 0 C,		\ 1 q# -> rax mov, 	;; no error
	     $FF C, $E3 C, 					\ rbx jmp,
	END-CODE
[ELSE]  CODE dummy ( addr -- u )
	     $B8 C,   1 C, 0 C, 0 C, 0 C, 			\ 1 d# -> eax mov, 	;; no error
	     $FF C, $E3 C, 					\ ebx jmp,
	END-CODE
[THEN]  PRIVATE

: SETUP-TEST ( -- )
	dll ABORT" Test already setup"
	S" kernel32.dll" LIBRARY-OPEN ABORT" LIBRARY-OPEN failed" TO dll
	CR ." Library handle = " dll H.
	S" GetNumberOfConsoleMouseButtons" dll LIBRARY-FIND
		ABORT" LIBRARY-FIND failed" TO '#buttons?
	S" GetTickCount" dll LIBRARY-FIND
		ABORT" LIBRARY-FIND failed" TO 'GTC
	CR ." Address of #BUTTONS? = " '#buttons? H.
	CR ." Address of GetTickCount = " 'GTC H.
	CR ." Latest error = " LIBRARY-ERROR  ?DUP IF TYPE EXIT
						 ELSE DROP ." <none>"
						ENDIF
	TICKS-RESET
	#1000000 0 ?DO  PAD 1 ['] dummy FOREIGN DROP \ 0= ABORT" dummy :: failed"
	          LOOP
	TICKS? TO 1e6calls
	PAD 1 CELLS ERASE ; PRIVATE

: CLEANUP-TEST ( -- )
	dll LIBRARY-CLOSE  CLEAR dll CLEAR '#buttons? CLEAR 'GTC ABORT" LIBRARY-CLOSE failed" ; PRIVATE

: OVERHEAD? ( n -- )
	0 LOCALS| gtc N |

	SETUP-TEST

	CR ." Executing " N DEC. ." * #BUTTONS? => "
	TICKS-RESET
	N 0 ?DO  PAD 1 '#buttons? FOREIGN 0= ABORT" #buttons? :: failed"
	   LOOP
        TICKS? N UM/MOD NIP DEC. ." ticks/call, overhead => " 1e6calls #1000000 UM/MOD NIP DEC. ." ticks."
	CR ." #buttons = " PAD ?

	CR ." Executing " N DEC. ." * GetTickCount => "
	TICKS-RESET
	N 0 ?DO  0 'GTC FOREIGN TO gtc  LOOP
        TICKS? N UM/MOD NIP DEC. ." ticks/call, overhead => " 1e6calls #1000000 UM/MOD NIP DEC. ." ticks."
	CR ." GetTickCount = " gtc U.

	CLEANUP-TEST ;
ENDDOC

DOC
(*
	FORTH> 1000 OVERHEAD? ( 166 MHz P54c on NT 4.0 )
	Library handle = $77F00000
	Address of #BUTTONS? = $77F29DF5
	Address of GetTickCount = $77F05FDF
	Latest error = <none>
	Executing 1000 * #BUTTONS? => 10953 ticks/call, overhead => 154 ticks.
	#buttons = 3
	Executing 1000 * GetTickCount => 145 ticks/call, overhead => 154 ticks.
	GetTickCount = 3855212642  ok

	FORTH> 100000 overhead? ( 900 MHz AMD Athlon on W2K )
	Library handle = $77E80000
	Address of #BUTTONS? = $77EC6B60
	Address of GetTickCount = $77E8C0A6
	Latest error = <none>
	Executing 100000 * #BUTTONS? => 6808 ticks/call, overhead => 63 ticks.
	#buttons = 3
	Executing 100000 * GetTickCount => 46 ticks/call, overhead => 63 ticks.
	GetTickCount = 310708184  ok
*)
ENDDOC


	NESTING @ 1 = [IF] .ABOUT -dynlibs CR [THEN]
	DEPRIVE

